/**
 * Copyright (C) 2003-2005 Funambol
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA 
 */

package sync4j.syncclient.test;

import java.io.*;
import java.security.Principal;
import java.util.Date;
import java.util.Vector;
import java.util.Enumeration;
import java.util.Properties;

import sync4j.syncclient.spds.engine.*;
import sync4j.syncclient.spds.SyncException;

import sync4j.framework.tools.Base64;


/**
 * This class implements a dummy <i>SyncSource</i> that just displays the calls
 * to its methods
 *
 * @author  Stefano Fornari
 *
 * @version $Id: FileSystemSyncSource.java,v 1.6.2.1 2005/01/27 11:00:57 fabius Exp $
 *
 */
public class FileSystemSyncSource implements SyncSource {

    // --------------------------------------------------------------- Constants

    public static final String DATABASE_FILE_NAME    = "sync.db";
    public static final String DATABASE_HEADER       = "FileSystemSyncSource file database";

    public static final String LOG_NAME = "sync4j.framework.engine";

    // -------------------------------------------------------------- Properties

    /**
     * The directory where files are stored (the default is the current
     * directory) - read/write
     */
    private String sourceDirectory = ".";

    public void setSourceDirectory(String sourceDirectory) {
        this.sourceDirectory = sourceDirectory;
    }

    public String getSourceDirectory() {
        return this.sourceDirectory;
    }

    private String name;

    public String getName()               {
        return name;
    }

    public void setName(String name)      {
        this.name = name;
    }

    private String type;

    public String getType()               {
        return this.type;
    }

    public void setType(String type)      {
        this.type = type;
    }

    private boolean encode = false;

    public boolean isEncode()             {
        return encode;
    }

    public void setEncode(boolean encode) {
        this.encode = encode;
    }

    public void setEncode(String encode)  {

        if ("true".equals(encode)) {
            this.encode = true  ;
        } else {
            this.encode = false ;
        }
    }

    private String sourceURI;

    /** Getter for property uri.
     * @return Value of property uri.
     */
    public String getSourceURI()          {
        return sourceURI;
    }

    /** Setter for property uri.
     * @param uri New value of property uri.
     */
    public void setSourceURI(String sourceURI) {
        this.sourceURI = sourceURI;
    }

    // ------------------------------------------------------------ Constructors

    /** Creates a new instance of AbstractSyncSource */
    public FileSystemSyncSource() {
    }

    // ---------------------------------------------------------- Public methods

    public void beginSync(int type) throws SyncException {
    	restoreSyncDB ();
    }


    public void endSync() throws SyncException {
    }


    public SyncItem[] getAllSyncItems(Principal principal) throws SyncException {
        return filterSyncItems(principal, null, SyncItemState.UNKNOWN);
    }


    public SyncItem[] getDeletedSyncItems(Principal principal,
                                          Date since    ) throws SyncException {
        return filterSyncItems(principal, since, SyncItemState.DELETED);

    }


    public SyncItem[] getNewSyncItems(Principal principal,
                                      Date since    ) throws SyncException {
        return filterSyncItems(principal, since, SyncItemState.NEW);

    }

    public SyncItem[] getUpdatedSyncItems(Principal principal,
                                          Date since    ) throws SyncException {
        return filterSyncItems(principal, since, SyncItemState.UPDATED);

    }

    public void removeSyncItem(Principal principal, SyncItem syncItem) throws SyncException {
        String fileName = syncItem.getKey().getKeyAsString();

        new File(sourceDirectory, fileName).delete();

        removeState(principal, fileName);
    }

    public SyncItem setSyncItem(Principal principal, SyncItem syncItem)
    throws SyncException {
        try {
            String fileName = syncItem.getKey().getKeyAsString();
            byte[] fileContent =
                (byte[])syncItem.getPropertyValue(SyncItem.PROPERTY_BINARY_CONTENT);

            FileOutputStream fos = new FileOutputStream(new File(sourceDirectory, fileName));
            if (fileContent != null) {
               if (encode) {
                   fos.write(Base64.decode(fileContent));
               } else {
                   fos.write(fileContent);
               }
            }
            fos.close();

            setState(principal, fileName, SyncItemState.SYNCHRONIZED);

            SyncItem newSyncItem =
                new SyncItemImpl(this, fileName, SyncItemState.NEW);

            newSyncItem.setProperties(syncItem.getProperties());

            return newSyncItem;
        } catch (IOException e) {
            throw new SyncException( "Error setting the item "
                                   + syncItem
                                   , e
                                   );
        }
    }

    public void commitSync() {
    }

    // ------------------------------------------------------------ Private data
    /**
     * The synchronization database is stored in a file whose name is given
     * by the value of the constant DATABASE_FILE_NAME prefixed by the user name
     * and the device id. For example:
     * <pre>
     *   guest.Sync4jTest.sync.db
     * </pre>
     * The database is a property file where each entry has the following
     * format:
     * <pre>
     * [filename]=[state][lastmodified_timestamp]
     * </pre>
     * For example:
     * <blockquote>
     * readme.txt=U98928743098094
     * </blockquote>
     * <p>
     * updateSyncDatabase works as follows:
     * <pre>
     * 1.    Read the existing database (if it exists)
     * 2.    Scan the source directory getting all files in the directory
     * 3.    For each file f in the source directory
     * 3.1.    If f is already in the database
     * 3.1.1.    If f has been modified after the lastmodified_timestamp stored into the database
     * 3.1.1.1.    Set the state of the file to UPDATE and store the new lastmodified_timestamp
     * 3.2.    Else
     * 3.2.1.    Add f to the database setting its state to NEW and store the lastmodified_timestamp
     * 3.    End For each
     * 4.    For each file f in the database
     * 4.1.    If f does not exist in the source directory
     * 4.1.1.    Set the state to DELETED
     * 5.    End For each
     * </pre>
     * <p>
     * At the end of the process, the updated database is saved and than returned.
     *
     * @return the updated databse. In case of error, an error message is traced
     *         and an empty Properties object is returned.
     */
    private Properties updateSyncDatabase(Principal principal) {

        Properties syncDatabase = new Properties();

        try {
            File fileSyncDatabase = getDatabaseFile(principal);

            //
            // Reads the existing database
            //
            if (fileSyncDatabase.exists()) {
                FileInputStream fis = new FileInputStream(fileSyncDatabase);
                syncDatabase.load(fis);
                fis.close();
            }

            //
            // Get the list of files in the source directory
            //
            Vector existingFiles = getExistingFiles();

            //
            // Get the list of the file in the databsae
            //
            Enumeration databaseFiles = syncDatabase.propertyNames();

            String state    = null,
                   fileName = null;

            long lastModified;

            int n = existingFiles.size();
            for (int i=0; i < n; ++i) {
                fileName = (String)existingFiles.elementAt(i);
                lastModified = new File(sourceDirectory, fileName).lastModified();

                state = syncDatabase.getProperty(fileName);

                if (state != null) {
                    //
                    // The file is already in the database
                    //
                    if (lastModified > lastModifiedFromStateString(state)) {
                        state = buildStateString(SyncItemState.UPDATED, lastModified);
                        syncDatabase.put(fileName, state);
                    }
                } else {
                    //
                    // The file is not in the database
                    //
                    state = buildStateString(SyncItemState.NEW, lastModified);
                    syncDatabase.put(fileName, state);
                }
            }  // next i

            for (; databaseFiles.hasMoreElements();) {
                fileName = (String)databaseFiles.nextElement();

                if (!existingFiles.contains(fileName)) {
                    state = buildStateString(SyncItemState.DELETED, System.currentTimeMillis());
                    syncDatabase.put(fileName, state);
                }
            }  // next

            //
            // Save & return
            //
            FileOutputStream fos = new FileOutputStream(fileSyncDatabase);
            syncDatabase.save(fos, DATABASE_HEADER);
            fos.close();
        } catch (IOException e) {
            e.printStackTrace();
        }

        return syncDatabase;
    }

    private File getDatabaseFile(Principal principal) {
        return new File(sourceDirectory + '.' + DATABASE_FILE_NAME);
    }

    private String buildStateString(char state, long lastModified) {
        return state + String.valueOf(lastModified);
    }

    private long lastModifiedFromStateString(String state) {
        return Long.parseLong(state.substring(1));
    }

    private char stateFromStateString(String state) {
        if ((state == null) || (state.length() == 0)) return SyncItemState.UNKNOWN;

        return state.charAt(0);
    }

    private Vector getExistingFiles() throws IOException {
        Vector ret = new Vector();

        String[] files = new File(sourceDirectory).list();

        if (files != null) {
            for (int i = 0; i<files.length; ++i) {
                if (!files[i].endsWith('.' + DATABASE_FILE_NAME)) {
                    ret.addElement(files[i]);
                }
            }  // next i
        }

        return ret;
    }

    private void setState(Principal principal, String file, char state) {
        try {
            Properties syncDatabase = new Properties();

            File fileSyncDatabase = getDatabaseFile(principal);

            //
            // Reads the existing database
            //
            if (fileSyncDatabase.exists()) {
                FileInputStream fis = new FileInputStream(fileSyncDatabase);
                syncDatabase.load(fis);
                fis.close();
            }

            syncDatabase.put(file, buildStateString(state, System.currentTimeMillis()));

            FileOutputStream fos = new FileOutputStream(fileSyncDatabase);
            syncDatabase.save(fos, DATABASE_HEADER);
            fos.close();
        } catch (IOException e) {
            e.printStackTrace();
        }
    }

    private void removeState(Principal principal, String file) {
        try {
            File fileSyncDatabase = getDatabaseFile(principal);

            if (!fileSyncDatabase.exists()) return;

            Properties syncDatabase = new Properties();

            //
            // Reads the existing database
            //

            FileInputStream fis = new FileInputStream(fileSyncDatabase);
            syncDatabase.load(fis);
            fis.close();

            syncDatabase.remove(file);

            FileOutputStream fos = new FileOutputStream(fileSyncDatabase);
            syncDatabase.save(fos, DATABASE_HEADER);
            fos.close();
        } catch (IOException e) {
            e.printStackTrace();
        }
    }

    /**
     * Filters the SyncItems in the synchronization database (after a refresh)
     * based on the given principal, last sync timestamp and state (see
     * SyncItemState). If state is equals to UNKNOWN all items are returned.<br>
     * Note that the current implementation ignores the principal: data do not
     * depend on users.
     *
     * @param principal principal. null means any
     * @param since     last sync timestamp. null neans since ever
     * @param state     the state to use as filter
     *
     * @return an array of SyncItem objects whose state is equal to the given
     *         state.
     */
    private SyncItem[] filterSyncItems(Principal principal,
                                       Date      since    ,
                                       char      state    ) {
        Properties syncDatabase = updateSyncDatabase(principal);

        Vector syncItems = new Vector();

        long fileTimestamp,
             sinceTimestamp = (since == null) ? -1 : since.getTime();

        SyncItem  syncItem      = null;
        String    fileName      = null;
        String    stateString   = null;
        char      fileState           ;
        for (Enumeration e = syncDatabase.keys(); e.hasMoreElements(); ) {
            fileName  = (String)e.nextElement();
            stateString = (String)syncDatabase.get(fileName);
            fileState = stateFromStateString(stateString);
            if ((state == SyncItemState.UNKNOWN) || (fileState == state)) {
                fileTimestamp = lastModifiedFromStateString(stateString);
                if (fileTimestamp > sinceTimestamp ) {
                    syncItem = new SyncItemImpl(this, fileName, fileState);
                    if (encode){
                        syncItem.setProperty(
                            new SyncItemProperty(SyncItem.PROPERTY_BINARY_CONTENT         ,
                                                 Base64.encode(readFileContent(fileName)))
                        );
                    } else {
                        syncItem.setProperty(
                            new SyncItemProperty(SyncItem.PROPERTY_BINARY_CONTENT,
                                                 readFileContent(fileName)       )
                        );
                    }
                    syncItems.addElement(syncItem);
                }
            }
        }  // next e

        SyncItem[] ret = new SyncItem[syncItems.size()];
        for (int i=0; i<ret.length; ++i) {
            ret[i] = (SyncItem)syncItems.elementAt(i);
        }

        return ret;
    }

    /**
     * Reads the content of the given file.
     *
     * @param   fileName    the name of the file to read
     *
     * @return the file content as a byte[]
     */
    private byte[] readFileContent(String fileName) {
        byte buf[] = null;

        try {
            File file = new File(sourceDirectory, fileName);

            if (!file.exists()) return new byte[0];

            buf = new byte[(int)file.length()];

            FileInputStream fis = new FileInputStream(file);
            fis.read(buf);
            fis.close();
        } catch (IOException e) {
            buf = new byte[0];
            e.printStackTrace();
        }

        return buf;
    }

	/**
	 * remove old deleted files from syncDB
	 */
	private void restoreSyncDB () {

		Properties syncDatabase = new Properties();

		String state    = null;
		String fileName = null;

		try {

			File fileSyncDatabase = getDatabaseFile(null);

			//
			// Reads the existing database
			//
			if (fileSyncDatabase.exists()) {
				FileInputStream fis = new FileInputStream(fileSyncDatabase);
				syncDatabase.load(fis);
				fis.close();
			}

			//
			// Get the list of the file in the databsae
			//
			Enumeration databaseFiles = syncDatabase.propertyNames();

			for (; databaseFiles.hasMoreElements();) {
				fileName = (String)databaseFiles.nextElement();

				state = syncDatabase.getProperty(fileName);

				if (stateFromStateString(state) == SyncItemState.DELETED) {
					syncDatabase.remove(fileName);
				}
			}  // next


			//
			// Save & return
			//
			FileOutputStream fos = new FileOutputStream(fileSyncDatabase);
			syncDatabase.save(fos, DATABASE_HEADER);
			fos.close();
		} catch (IOException e) {
			e.printStackTrace();
		}

	}


}